#include <OneWire.h>
#include <DallasTemperature.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <FS.h>

#define ONE_HOUR 3600000UL

#define CONFIG_HTTP_PORT                          80
#define CONFIG_UDP_PORT                           123
#define CONFIG_WIFI_SSID                          "OpenWrt"
#define CONFIG_WIFI_PSK                           "3JGJXC3R"
#define CONFIG_NTP_UPDATE_INTERVAL                ONE_HOUR      // Update the time every hour
#define CONFIG_TEMPERATURE_SENSOR_PIN             5
#define CONFIG_TEMPERATURE_MEASUREMENT_INTERVAL   60000         // Do a temperature measurement every minute

OneWire oneWire(CONFIG_TEMPERATURE_SENSOR_PIN); // Set up a OneWire instance to communicate with OneWire devices
DallasTemperature tempSensors(&oneWire);        // Create an instance of the temperature sensor class
ESP8266WebServer server(CONFIG_HTTP_PORT);      // create a web server on port 80
File fsUploadFile;                              // a File variable to temporarily store the received file
ESP8266WiFiMulti wifiMulti;                     // Create an instance of the ESP8266WiFiMulti class, called 'wifiMulti'
const char *OTAName = "";                       // A name and a password for the OTA service
const char *OTAPassword = "";
const char* mdnsName = "tempsensor1";           // Domain name for the mDNS responder
WiFiUDP UDP;                                    // Create an instance of the WiFiUDP class to send and receive UDP messages
IPAddress timeServerIP;
const char* ntpServerName = "time.google.com";  // The NTP server's IP address
const int NTP_PACKET_SIZE = 48;                 // NTP time stamp is in the first 48 bytes of the message
byte packetBuffer[NTP_PACKET_SIZE];             // A buffer to hold incoming and outgoing packets

/* SETUP
 * ================================================================================================================== */

void setup()
{
    Serial.begin(115200); // Start the Serial communication to send messages to the computer
    delay(10);
    Serial.println("\r\n");

    tempSensors.setWaitForConversion(false); // Don't block the program while the temperature sensor is reading
    tempSensors.begin();                     // Start the temperature sensor

    if (tempSensors.getDeviceCount() == 0) {
        Serial.printf("No DS18x20 temperature sensor found on pin %d. Rebooting.\r\n", CONFIG_TEMPERATURE_SENSOR_PIN);
        Serial.flush();
        ESP.reset();
    }

    ESP_StartWiFi();       // Start a Wi-Fi access point, and try to connect to some given access points. Then wait for either an AP or STA connection
    ESP_StartOTA();        // Start the OTA service
    ESP_StartSPIFFS();     // Start the SPIFFS and list all contents
    ESP_StartMDNS();       // Start the mDNS responder
    ESP_StartHTTPServer(); // Start a HTTP server with a file read handler and an upload handler
    ESP_StartUDP();        // Start listening for UDP messages to port 123

    WiFi.hostByName(ntpServerName, timeServerIP); // Get the IP address of the NTP server
    Serial.print("Time server IP:\t");
    Serial.println(timeServerIP);

    sendNTPpacket(timeServerIP);
    delay(500);
}

/* LOOP
 * ================================================================================================================== */

const uint32_t intervalNTP = CONFIG_NTP_UPDATE_INTERVAL;
uint32_t prevNTP = 0;
uint32_t lastNTPResponse = millis();

const uint32_t intervalTemp = CONFIG_TEMPERATURE_MEASUREMENT_INTERVAL;
uint32_t prevTemp = 0;
bool tmpRequested = false;
const uint32_t DS_delay = 750;         // Reading the temperature from the DS18x20 can take up to 750ms

uint32_t timeUNIX = 0;                      // The most recent timestamp received from the time server

void loop()
{
    uint32_t currentMillis = millis();

    if (currentMillis - prevNTP > intervalNTP) { // Request the time from the time server every hour
        prevNTP = currentMillis;
        sendNTPpacket(timeServerIP);
    }

    uint32_t time = getTime();                   // Check if the time server has responded, if so, get the UNIX time
    if (time) {
        timeUNIX = time;
        Serial.print("NTP response:\t");
        Serial.println(timeUNIX);
        lastNTPResponse = millis();
    } else if ((millis() - lastNTPResponse) > (24UL * ONE_HOUR)) {
        Serial.println("More than 24 hours since last NTP response. Rebooting.");
        Serial.flush();
        ESP.reset();
    }

    if (timeUNIX == 0) {
        // if we didn't receive an NTP response yet, send another request
        sendNTPpacket(timeServerIP);
        delay(500);
    } else {
        // Every minute, request the temperature
        if (currentMillis - prevTemp > intervalTemp)
        {
            tempSensors.requestTemperatures(); // Request the temperature from the sensor (it takes some time to read it)
            tmpRequested = true;
            prevTemp = currentMillis;
            Serial.println("Temperature requested");
        }

        // 750 ms after requesting the temperature
        if (currentMillis - prevTemp > DS_delay && tmpRequested)
        {
            // The actual time is the last NTP time plus the time that has elapsed since the last NTP response
            uint32_t actualTime = timeUNIX + (currentMillis - lastNTPResponse) / 1000;
            tmpRequested = false;

            float temp = tempSensors.getTempCByIndex(0); // Get the temperature from the sensor
//            temp = round(temp * 100.0) / 100.0; // round temperature to 2 digits

            Serial.printf("Appending temperature to file: %lu,\n", actualTime);

            File tempLog = SPIFFS.open("/data.csv", "a"); // Write the time and the temperature to the csv file
            if (tempLog) {
                tempLog.printf("%lu,%.2f\n", actualTime, temp);
                tempLog.close();
            }
        }
    }

    server.handleClient(); // run the server
    ArduinoOTA.handle();   // listen for OTA events
    MDNS.update();
}

/* SETUP FUNCTIONS
 * ================================================================================================================== */

// Try to connect to some given access points. Then wait for a connection
void ESP_StartWiFi()
{
    wifiMulti.addAP(CONFIG_WIFI_SSID, CONFIG_WIFI_PSK); // add Wi-Fi networks you want to connect to

    Serial.println("Connecting");
    while (wifiMulti.run() != WL_CONNECTED) { // Wait for the Wi-Fi to connect
        delay(250);
        Serial.print('.');
    }
    Serial.println("\r\n");
    Serial.print("Connected to ");
    Serial.println(WiFi.SSID());             // Tell us what network we're connected to
    Serial.print("IP address:\t");
    Serial.print(WiFi.localIP());            // Send the IP address of the ESP8266 to the computer
    Serial.println("\r\n");
}

// Start listening for UDP messages
void ESP_StartUDP()
{
    Serial.println("Starting UDP");
    UDP.begin(CONFIG_UDP_PORT);
    Serial.print("Local port:\t");
    Serial.println(UDP.localPort());
}

// Start the OTA service
void ESP_StartOTA()
{
    ArduinoOTA.setHostname(OTAName);
    ArduinoOTA.setPassword(OTAPassword);

    ArduinoOTA.onStart([]() {
        Serial.println("Start");
    });
    ArduinoOTA.onEnd([]() {
        Serial.println("\r\nEnd");
    });
    ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
        Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    });
    ArduinoOTA.onError([](ota_error_t error) {
        Serial.printf("Error[%u]: ", error);
             if (error == OTA_AUTH_ERROR)    { Serial.println("Auth Failed"); }
        else if (error == OTA_BEGIN_ERROR)   { Serial.println("Begin Failed"); }
        else if (error == OTA_CONNECT_ERROR) { Serial.println("Connect Failed"); }
        else if (error == OTA_RECEIVE_ERROR) { Serial.println("Receive Failed"); }
        else if (error == OTA_END_ERROR)     { Serial.println("End Failed"); }
    });
    ArduinoOTA.begin();
    Serial.println("OTA ready\r\n");
}

// Start the SPIFFS and list all contents
void ESP_StartSPIFFS()
{
    if (SPIFFS.begin() == true) { // Start the SPI Flash File System (SPIFFS)

        // List the file system contents
        Serial.println("SPIFFS started. Contents:");
        Dir dir = SPIFFS.openDir("/");
        while (dir.next()) {
            String fileName = dir.fileName();
            size_t fileSize = dir.fileSize();
            Serial.printf("\tFS File: %s, size: %s\r\n", fileName.c_str(), formatBytes(fileSize).c_str());
        }
        Serial.printf("\n");
    }
}

// Start the mDNS responder
void ESP_StartMDNS()
{
    if (MDNS.begin(mdnsName) == true) { // start the multicast domain name server
        Serial.printf("mDNS responder started: http://%s.local\n", mdnsName);
    }
}

// Start a HTTP server with a file read handler and an upload handler
void ESP_StartHTTPServer()
{
    server.on("/api/file", HTTP_GET, handleFileList);
    server.on("/api/file", HTTP_POST, []() {    // if a POST request is sent to the /edit.html address,
        server.send(200, "text/plain", "");
    }, handleFileUpload);                       // go to 'handleFileUpload'
    server.on("/api/file", HTTP_DELETE, handleFileDelete);
    server.on("/api/file", HTTP_PUT, handleFileCreate);

    server.onNotFound(handleNotFound);          // if someone requests any other file or page, go to function 'handleNotFound'
                                                // and check if the file exists

    server.begin();                             // start the HTTP server
    Serial.println("HTTP server started.");
}

/* SERVER HANDLERS
 * ================================================================================================================== */

// if the requested file or page doesn't exist, return a 404 not found error
void handleNotFound()
{
    if (handleFileRead(server.uri()) == false) { // check if the file exists in the flash memory (SPIFFS), if so, send it
        server.send(404, "text/plain", "404 Not Found");
    }
}

// send the right file to the client (if it exists)
bool handleFileRead(String path)
{
    Serial.println("handleFileRead: " + path);

    // If a folder is requested, send the index file
    if (path.endsWith("/")) {
        path += "index.html";
    }

    String pathWithGz = path + ".gz";

    bool isPathExists       = SPIFFS.exists(path);
    bool isPathWithGzExists = SPIFFS.exists(pathWithGz);  

    if (isPathWithGzExists || isPathExists) { // If the file exists, either as a compressed archive, or normal
    
        String contentType = getContentType(path); // Get the MIME type
      
        if (isPathWithGzExists) { // If there's a compressed version available
            path = pathWithGz;    // Use the compressed verion
        }

        File file = SPIFFS.open(path, "r");
        if (file) {
            size_t sent = server.streamFile(file, contentType); // Send it to the client
            file.close();

            Serial.printf("\tSent file: %s (%lu bytes)\n", path.c_str(), sent);
            return true;
        }
    }
    Serial.printf("\tFile Not Found: %s\n", path.c_str()); // If the file doesn't exist, return false
    return false;
}

// upload a new file to the SPIFFS
void handleFileUpload()
{
    HTTPUpload& upload = server.upload();
    String path;
    if (upload.status == UPLOAD_FILE_START) {
        path = upload.filename;
        if (!path.startsWith("/")) path = "/" + path;
        if (!path.endsWith(".gz")) {                           // The file server always prefers a compressed version of a file
            String pathWithGz = path + ".gz";                  // So if an uploaded file is not compressed, the existing compressed
            if (SPIFFS.exists(pathWithGz))                     // version of that file must be deleted (if it exists)
                SPIFFS.remove(pathWithGz);
        }
        Serial.print("handleFileUpload() Name: "); Serial.println(path);
        fsUploadFile = SPIFFS.open(path, "w");                 // Open the file for writing in SPIFFS (create if it doesn't exist)
        path = String();
    } else if (upload.status == UPLOAD_FILE_WRITE) {
        if (fsUploadFile) {
            fsUploadFile.write(upload.buf, upload.currentSize); // Write the received bytes to the file
        }
    } else if (upload.status == UPLOAD_FILE_END) {
        if (fsUploadFile) {                                     // If the file was successfully created
            fsUploadFile.close();                               // Close the file again
            Serial.print("handleFileUpload() Size: "); Serial.println(upload.totalSize);
            server.sendHeader("Location", "/files.html");     // Redirect the client to the success page
            server.send(303);
        } else {
            server.send(500, "text/plain", "500 Couldn't create file");
        }
    }
}

void handleFileDelete()
{
  if (server.args() == 0) {
    return server.send(500, "text/plain", "Bad request");
  }
  String path = server.arg(0);
  Serial.println("handleFileDelete: " + path);
  if (path == "/") {
    return server.send(500, "text/plain", "Bad path");
  }
  if (!SPIFFS.exists(path)) {
    return server.send(404, "text/plain", "FileNotFound");
  }
  SPIFFS.remove(path);
  server.send(200, "text/plain", "");
  path = String();
}

void handleFileCreate()
{
  if (server.args() == 0) {
    return server.send(500, "text/plain", "Bad request");
  }
  String path = server.arg(0);
  Serial.println("handleFileCreate: " + path);
  if (path == "/") {
    return server.send(500, "text/plain", "Bad path");
  }
  if (SPIFFS.exists(path)) {
    return server.send(500, "text/plain", "File exists");
  }
  File file = SPIFFS.open(path, "w");
  if (file) {
    file.close();
  } else {
    return server.send(500, "text/plain", "Create failed");
  }
  server.send(200, "text/plain", "");
  path = String();
}

void handleFileList()
{
    String path = server.hasArg("dir") ? server.arg("dir") : "/";
    Serial.println("handleFileList: " + path);
    Dir dir = SPIFFS.openDir(path);

    String output = "";
    while (dir.next()) {
        output += dir.isDirectory();
        output += ",";
        output += dir.fileName();
        output += ",";
        output += dir.fileSize();
        output += "\n";
    }

    server.send(200, "text/plain", output);
}


/* HELPER FUNCTIONS
 * ================================================================================================================== */

// convert sizes in bytes to KB and MB
String formatBytes(size_t bytes)
{
         if (bytes <  1024)                { return String(bytes)                   +  "B"; }
    else if (bytes < (1024 * 1024))        { return String(bytes / 1024.0)          + "KB"; }
    else if (bytes < (1024 * 1024 * 1024)) { return String(bytes / 1024.0 / 1024.0) + "MB"; }
}

// determine the filetype of a given filename, based on the extension
String getContentType(String filename)
{
         if (filename.endsWith(".html")) { return "text/html"; }
    else if (filename.endsWith(".css"))  { return "text/css"; }
    else if (filename.endsWith(".js"))   { return "application/javascript"; }
    else if (filename.endsWith(".ico"))  { return "image/x-icon"; }
    else if (filename.endsWith(".gz"))   { return "application/x-gzip"; }
    else if (filename.endsWith(".png"))  { return "image/png"; }
    else if (filename.endsWith(".gif"))  { return "image/gif"; }
    else if (filename.endsWith(".jpg"))  { return "image/jpeg"; }
    else if (filename.endsWith(".ico"))  { return "image/x-icon"; }
    else if (filename.endsWith(".xml"))  { return "text/xml"; }
    
    return "text/plain";
}

// Check if the time server has responded, if so, get the UNIX time, otherwise, return 0
uint32_t getTime()
{
    if (UDP.parsePacket() == 0) { // if there's no response (yet)
        return 0;
    }

    UDP.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer

    // Combine the 4 timestamp bytes into one 32-bit number
    uint32_t NTPTime =   (packetBuffer[40] << 24)
                       | (packetBuffer[41] << 16)
                       | (packetBuffer[42] <<  8)
                       |  packetBuffer[43];

    // Convert NTP time to a UNIX timestamp:
    // Unix time starts on Jan 1 1970. That's 2208988800 seconds in NTP time:
    const uint32_t SEVENTY_YEARS = 2208988800UL;

    // subtract seventy years:
    uint32_t UNIXTime = NTPTime - SEVENTY_YEARS;
    return   UNIXTime;
}

void sendNTPpacket(IPAddress& address)
{
    Serial.println("Sending NTP request");
    memset(packetBuffer, 0, NTP_PACKET_SIZE); // set all bytes in the buffer to 0
    // Initialize values needed to form NTP request
    packetBuffer[0] = 0b11100011;   // LI, Version, Mode

    // send a packet requesting a timestamp:
    UDP.beginPacket(address, 123); // NTP requests are to port 123
    UDP.write(packetBuffer, NTP_PACKET_SIZE);
    UDP.endPacket();
}
